"""Generate the test code for the JSON de/serialization of enums."""

import io
import os
import pathlib
import sys
from typing import List

import aas_core_codegen
import aas_core_codegen.common
import aas_core_codegen.naming
import aas_core_codegen.parse
import aas_core_codegen.python.common
import aas_core_codegen.python.naming
import aas_core_codegen.run
from aas_core_codegen import intermediate
from aas_core_codegen.common import Stripped
from aas_core_codegen.python.common import INDENT as I, INDENT2 as II, INDENT3 as III

import dev_scripts.test_codegen.common


def main() -> int:
    """Execute the main routine."""
    symbol_table = dev_scripts.test_codegen.common.load_symbol_table()

    aas_module = dev_scripts.test_codegen.common.AAS_MODULE

    this_path = pathlib.Path(os.path.realpath(__file__))

    warning = dev_scripts.test_codegen.common.generate_warning_comment(
        this_path.relative_to(dev_scripts.test_codegen.common.REPO_ROOT)
    )

    blocks = [
        warning,
        Stripped('"""Test JSON de/serialization of enumeration literals."""'),
        Stripped("# pylint: disable=missing-docstring"),
        Stripped(
            f"""\
import unittest

import {aas_module}.jsonization as aas_jsonization"""
        ),
    ]  # type: List[Stripped]

    for our_type in symbol_table.our_types:
        if not isinstance(our_type, intermediate.Enumeration):
            continue

        enum_name = aas_core_codegen.python.naming.enum_name(our_type.name)

        assert (
            len(our_type.literals) > 0
        ), f"Unexpected enumeration without literals: {our_type.name}"

        literal_value = our_type.literals[0].value

        from_jsonable_name = aas_core_codegen.python.naming.function_name(
            aas_core_codegen.common.Identifier(f"{our_type.name}_from_jsonable")
        )

        blocks.append(
            Stripped(
                f"""\
class Test_{enum_name}(unittest.TestCase):
{I}def test_round_trip(self) -> None:
{II}jsonable = {aas_core_codegen.python.common.string_literal(literal_value)}

{II}enum_literal = aas_jsonization.{from_jsonable_name}(
{III}jsonable
{II})

{II}self.assertEqual(enum_literal.value, jsonable)"""
            )
        )

    blocks.append(
        Stripped(
            """\
if __name__ == "__main__":
    unittest.main()"""
        )
    )

    blocks.append(warning)

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n\n")

        writer.write(block)

    writer.write("\n")

    target_pth = (
        dev_scripts.test_codegen.common.REPO_ROOT / "tests/test_jsonization_of_enums.py"
    )
    target_pth.write_text(writer.getvalue(), encoding="utf-8")

    return 0


if __name__ == "__main__":
    sys.exit(main())



"""Generate the test code for the JSON de/serialization of classes."""

import io
import os
import pathlib
import sys
import textwrap
from typing import List

import aas_core_codegen
import aas_core_codegen.common
import aas_core_codegen.naming
import aas_core_codegen.parse
import aas_core_codegen.run
from aas_core_codegen import intermediate
from aas_core_codegen.common import Stripped
import aas_core_codegen.python.naming
import aas_core_codegen.python.common
from aas_core_codegen.python.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
    INDENT5 as IIIII,
)
from icontract import require

import dev_scripts.test_codegen.common
import dev_scripts.test_codegen.test_data_io


def _generate_for_self_contained(
    cls: intermediate.ConcreteClass,
    aas_module: aas_core_codegen.python.common.QualifiedModuleName,
) -> Stripped:
    """Generate the test class for a self-contained class."""
    # noinspection PyListCreation
    cls_name_json_literal = aas_core_codegen.python.common.string_literal(
        aas_core_codegen.naming.json_model_type(cls.name)
    )

    deserialization_function = aas_core_codegen.python.naming.function_name(
        aas_core_codegen.common.Identifier(f"{cls.name}_from_jsonable")
    )

    cls_name = aas_core_codegen.python.naming.class_name(cls.name)

    methods = [
        Stripped(
            f"""\
def test_ok(self) -> None:
{I}paths = sorted(
{II}(
{IIII}tests.common.TEST_DATA_DIR
{IIII}/ "Json"
{IIII}/ "SelfContained"
{IIII}/ "Expected"
{IIII}/ {cls_name_json_literal}
{II}).glob("**/*.json")
{I})

{I}for path in paths:
{II}with path.open("rt") as fid:
{III}jsonable = json.load(fid)

{II}instance = aas_jsonization.{deserialization_function}(jsonable)

{II}errors = list(aas_verification.verify(instance))

{II}self.assertListEqual([], errors, f"path is: {{path}}")

{II}# Check the round-trip
{II}another_jsonable = aas_jsonization.to_jsonable(instance)
{II}self.assertListEqual(
{III}[],
{III}list(
{IIII}map(
{IIIII}str,
{IIIII}tests.common_jsonization.check_equal(jsonable, another_jsonable)
{IIII})
{III})
{II})"""
        ),
        Stripped(
            f"""\
def test_that_deserialization_from_non_object_fails(self) -> None:
{I}jsonable = "not an object"

{I}observed_exception: Optional[
{II}aas_jsonization.DeserializationException
{I}] = None

{I}try:
{II}aas_jsonization.{deserialization_function}(jsonable)
{I}except aas_jsonization.DeserializationException as exception:
{II}observed_exception = exception

{I}assert observed_exception is not None
{I}self.assertEqual(
{II}"Expected a mapping, but got: <class 'str'>",
{II}observed_exception.cause
{I})

{I}self.assertEqual("", str(observed_exception.path))"""
        ),
        Stripped(
            f"""\
def test_deserialization_failures(self) -> None:
{I}for cause in _CAUSES_FOR_DESERIALIZATION_FAILURE:
{II}base_dir = (
{III}tests.common.TEST_DATA_DIR
{III}/ "Json"
{III}/ "SelfContained"
{III}/ "Unexpected"
{III}/ cause
{III}/ {cls_name_json_literal}
{II})

{II}if not base_dir.exists():
{III}# There are no failure cases
{III}# for :py:class:`{aas_module}.types.{cls_name}`
{III}# and this ``cause``.
{III}continue

{II}for path in sorted(base_dir.glob("**/*.json")):
{III}with path.open("rt") as fid:
{IIII}jsonable = json.load(fid)

{III}observed_exception: Optional[
{IIII}aas_jsonization.DeserializationException
{III}] = None

{III}try:
{IIII}_ = aas_jsonization.{deserialization_function}(jsonable)
{III}except aas_jsonization.DeserializationException as exception:
{IIII}observed_exception = exception

{III}assert observed_exception is not None, (
{IIII}f"Expected an exception, but got none for: {{path}}"
{III})

{III}tests.common.record_or_check(
{IIII}path=path.parent / (path.name + ".exception"),
{IIII}got=f"{{observed_exception.path}}: {{observed_exception.cause}}"
{III})"""
        ),
        Stripped(
            f"""\
def test_verification_failures(self) -> None:
{I}for cause in tests.common.CAUSES_FOR_VERIFICATION_FAILURE:
{II}base_dir = (
{III}tests.common.TEST_DATA_DIR
{III}/ "Json"
{III}/ "SelfContained"
{III}/ "Unexpected"
{III}/ cause
{III}/ {cls_name_json_literal}
{II})

{II}if not base_dir.exists():
{III}# There are no failure cases
{III}# for :py:class:`{aas_module}.types.{cls_name}`
{III}# and this ``cause``.
{III}continue

{II}for path in sorted(base_dir.glob("**/*.json")):
{III}with path.open("rt") as fid:
{IIII}jsonable = json.load(fid)

{III}try:
{IIII}instance = aas_jsonization.{deserialization_function}(jsonable)
{III}except aas_jsonization.DeserializationException as exception:
{IIII}raise AssertionError(
{IIIII}f"Expected no deserialization exception from {{path}}"
{IIII}) from exception

{III}errors = list(aas_verification.verify(instance))

{III}self.assertGreater(
{IIII}len(errors),
{IIII}0,
{IIII}f"Expected verification errors from {{path}}, but got none"
{III})

{III}tests.common.record_or_check(
{IIII}path=path.parent / (path.name + ".errors"),
{IIII}got="\\n".join(
{IIIII}f"{{error.path}}: {{error.cause}}"
{IIIII}for error in errors
{IIII})
{III})"""
        ),
    ]

    writer = io.StringIO()
    writer.write(
        f"""\
class Test_{cls_name}(unittest.TestCase):
{I}\"\"\"
{I}Test JSON de/serialization of the concrete class
{I}:py:class:`{aas_module}.types.{cls_name}`.
{I}\"\"\""""
    )

    for method in methods:
        writer.write("\n\n")
        writer.write(textwrap.indent(method, I))

    return Stripped(writer.getvalue())


# fmt: off
@require(
    lambda container_cls:
    container_cls.name == "Environment",
    "Hint for the future reader what to expect; "
    "assume this might change in the future"
)
# fmt: on
def _generate_for_contained_in_environment(
    cls: intermediate.ConcreteClass,
    container_cls: intermediate.ConcreteClass,
    aas_module: aas_core_codegen.python.common.QualifiedModuleName,
) -> Stripped:
    """Generate the tests for a class contained in an ``Environment`` instance."""
    cls_name_json_literal = aas_core_codegen.python.common.string_literal(
        aas_core_codegen.naming.json_model_type(cls.name)
    )

    deserialization_function = aas_core_codegen.python.naming.function_name(
        aas_core_codegen.common.Identifier(f"{container_cls.name}_from_jsonable")
    )

    cls_name = aas_core_codegen.python.naming.class_name(cls.name)

    methods = [
        Stripped(
            f"""\
def test_ok(self) -> None:
{I}paths = sorted(
{II}(
{IIII}tests.common.TEST_DATA_DIR
{IIII}/ "Json"
{IIII}/ "ContainedInEnvironment"
{IIII}/ "Expected"
{IIII}/ {cls_name_json_literal}
{II}).glob("**/*.json")
{I})

{I}for path in paths:
{II}with path.open("rt") as fid:
{III}jsonable = json.load(fid)

{II}container = aas_jsonization.{deserialization_function}(jsonable)

{II}errors = list(aas_verification.verify(container))

{II}self.assertListEqual([], errors, f"path is {{path}}")

{II}# Check the round-trip
{II}another_jsonable = aas_jsonization.to_jsonable(container)
{II}self.assertListEqual(
{III}[],
{III}list(
{IIII}map(
{IIIII}str,
{IIIII}tests.common_jsonization.check_equal(jsonable, another_jsonable)
{IIII})
{III})
{II})"""
        ),
        Stripped(
            f"""\
def test_that_deserialization_from_non_object_fails(self) -> None:
{I}jsonable = "not an object"

{I}observed_exception: Optional[
{II}aas_jsonization.DeserializationException
{I}] = None

{I}try:
{II}aas_jsonization.{deserialization_function}(jsonable)
{I}except aas_jsonization.DeserializationException as exception:
{II}observed_exception = exception

{I}assert observed_exception is not None
{I}self.assertEqual(
{II}"Expected a mapping, but got: <class 'str'>",
{II}observed_exception.cause
{I})

{I}self.assertEqual("", str(observed_exception.path))"""
        ),
        Stripped(
            f"""\
def test_deserialization_failures(self) -> None:
{I}for cause in _CAUSES_FOR_DESERIALIZATION_FAILURE:
{II}base_dir = (
{III}tests.common.TEST_DATA_DIR
{III}/ "Json"
{III}/ "ContainedInEnvironment"
{III}/ "Unexpected"
{III}/ cause
{III}/ {cls_name_json_literal}
{II})

{II}if not base_dir.exists():
{III}# There are no failure cases
{III}# for :py:class:`{aas_module}.types.{cls_name}`
{III}# and this ``cause``.
{III}continue

{II}for path in sorted(base_dir.glob("**/*.json")):
{III}with path.open("rt") as fid:
{IIII}jsonable = json.load(fid)

{III}observed_exception: Optional[
{IIII}aas_jsonization.DeserializationException
{III}] = None

{III}try:
{IIII}_ = aas_jsonization.{deserialization_function}(jsonable)
{III}except aas_jsonization.DeserializationException as exception:
{IIII}observed_exception = exception

{III}assert observed_exception is not None, (
{IIII}f"Expected an exception, but got none for: {{path}}"
{III})

{III}tests.common.record_or_check(
{IIII}path=path.parent / (path.name + ".exception"),
{IIII}got=f"{{observed_exception.path}}: {{observed_exception.cause}}"
{III})"""
        ),
        Stripped(
            f"""\
def test_verification_failures(self) -> None:
{I}for cause in tests.common.CAUSES_FOR_VERIFICATION_FAILURE:
{II}base_dir = (
{III}tests.common.TEST_DATA_DIR
{III}/ "Json"
{III}/ "ContainedInEnvironment"
{III}/ "Unexpected"
{III}/ cause
{III}/ {cls_name_json_literal}
{II})

{II}if not base_dir.exists():
{III}# There are no failure cases
{III}# for :py:class:`{aas_module}.types.{cls_name}`
{III}# and this ``cause``.
{III}continue

{II}for path in sorted(base_dir.glob("**/*.json")):
{III}with path.open("rt") as fid:
{IIII}jsonable = json.load(fid)

{III}try:
{IIII}container = aas_jsonization.{deserialization_function}(jsonable)
{III}except aas_jsonization.DeserializationException as exception:
{IIII}raise AssertionError(
{IIIII}f"Expected no deserialization exception from {{path}}"
{IIII}) from exception

{III}errors = list(aas_verification.verify(container))

{III}self.assertGreater(
{IIII}len(errors),
{IIII}0,
{IIII}f"Expected verification errors from {{path}}, but got none"
{III})

{III}tests.common.record_or_check(
{IIII}path=path.parent / (path.name + ".errors"),
{IIII}got="\\n".join(
{IIIII}f"{{error.path}}: {{error.cause}}"
{IIIII}for error in errors
{IIII})
{III})"""
        ),
    ]

    writer = io.StringIO()
    writer.write(
        f"""\
class Test_{cls_name}(unittest.TestCase):
{I}\"\"\"
{I}Test JSON de/serialization of the concrete class
{I}:py:class:`{aas_module}.types.{cls_name}`.
{I}\"\"\""""
    )

    for method in methods:
        writer.write("\n\n")
        writer.write(textwrap.indent(method, I))

    return Stripped(writer.getvalue())


def main() -> int:
    """Execute the main routine."""
    symbol_table = dev_scripts.test_codegen.common.load_symbol_table()

    this_path = pathlib.Path(os.path.realpath(__file__))

    test_data_dir = dev_scripts.test_codegen.common.REPO_ROOT / "test_data"

    warning = dev_scripts.test_codegen.common.generate_warning_comment(
        this_path.relative_to(dev_scripts.test_codegen.common.REPO_ROOT)
    )

    aas_module = dev_scripts.test_codegen.common.AAS_MODULE

    blocks = [
        warning,
        Stripped('"""Test JSON de/serialization of concrete classes."""'),
        Stripped("# pylint: disable=missing-docstring"),
        Stripped(
            f"""\
import json
import unittest
from typing import Optional

import {aas_module}.jsonization as aas_jsonization
import {aas_module}.verification as aas_verification

import tests.common
import tests.common_jsonization"""
        ),
        Stripped(
            f"""\
#: List of identifiers for serialization failures (corresponding to subdirectory names
#: in the directory with test data)
_CAUSES_FOR_DESERIALIZATION_FAILURE = [
{I}"TypeViolation",
{I}"RequiredViolation",
{I}"EnumViolation",
{I}"NullViolation",
{I}"UnexpectedAdditionalProperty",
]"""
        ),
    ]  # type: List[Stripped]

    environment_cls = symbol_table.must_find_concrete_class(
        aas_core_codegen.common.Identifier("Environment")
    )

    for our_type in symbol_table.our_types:
        if not isinstance(our_type, intermediate.ConcreteClass):
            continue

        container_cls = dev_scripts.test_codegen.test_data_io.determine_container_class(
            cls=our_type, test_data_dir=test_data_dir, environment_cls=environment_cls
        )

        if container_cls is our_type:
            blocks.append(
                _generate_for_self_contained(
                    cls=our_type, aas_module=dev_scripts.test_codegen.common.AAS_MODULE
                )
            )
        else:
            blocks.append(
                _generate_for_contained_in_environment(
                    cls=our_type,
                    container_cls=container_cls,
                    aas_module=dev_scripts.test_codegen.common.AAS_MODULE,
                )
            )

    blocks.append(
        Stripped(
            """\
if __name__ == "__main__":
    unittest.main()"""
        )
    )

    blocks.append(warning)

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n\n")

        writer.write(block)

    writer.write("\n")

    target_pth = (
        dev_scripts.test_codegen.common.REPO_ROOT
        / "tests/test_jsonization_of_concrete_classes.py"
    )
    target_pth.write_text(writer.getvalue(), encoding="utf-8")

    return 0


if __name__ == "__main__":
    sys.exit(main())



"""Generate the test code for the XML de/serialization of classes with descendants."""

import io
import os
import pathlib
import sys
import textwrap
from typing import List

import aas_core_codegen
import aas_core_codegen.common
import aas_core_codegen.naming
import aas_core_codegen.parse
import aas_core_codegen.python.common
import aas_core_codegen.python.naming
import aas_core_codegen.run
from aas_core_codegen import intermediate
from aas_core_codegen.common import Stripped
from aas_core_codegen.python.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
)

import dev_scripts.test_codegen.common


def main() -> int:
    """Execute the main routine."""
    symbol_table = dev_scripts.test_codegen.common.load_symbol_table()

    this_path = pathlib.Path(os.path.realpath(__file__))

    warning = dev_scripts.test_codegen.common.generate_warning_comment(
        this_path.relative_to(dev_scripts.test_codegen.common.REPO_ROOT)
    )

    aas_module = dev_scripts.test_codegen.common.AAS_MODULE

    blocks = [
        warning,
        Stripped('"""Test XML de/serialization of classes with descendants."""'),
        Stripped("# pylint: disable=missing-docstring"),
        Stripped(
            f"""\
import io
import pathlib
import tempfile
import unittest
import xml.etree.ElementTree as ET

import {aas_module}.xmlization as aas_xmlization

import tests.common
import tests.common_jsonization
import tests.common_xmlization"""
        ),
    ]  # type: List[Stripped]

    for our_type in symbol_table.our_types:
        if not isinstance(
            our_type, (intermediate.AbstractClass, intermediate.ConcreteClass)
        ):
            continue

        if len(our_type.concrete_descendants) == 0:
            continue

        cls_name = aas_core_codegen.python.naming.class_name(our_type.name)

        test_methods = []  # type: List[Stripped]
        for descendant in our_type.concrete_descendants:
            load_concrete_maximal_name = aas_core_codegen.python.naming.function_name(
                aas_core_codegen.common.Identifier(f"load_maximal_{descendant.name}")
            )

            from_iterparse = aas_core_codegen.python.naming.function_name(
                aas_core_codegen.common.Identifier(f"{our_type.name}_from_iterparse")
            )

            from_stream = aas_core_codegen.python.naming.function_name(
                aas_core_codegen.common.Identifier(f"{our_type.name}_from_stream")
            )

            from_file = aas_core_codegen.python.naming.function_name(
                aas_core_codegen.common.Identifier(f"{our_type.name}_from_file")
            )

            from_str = aas_core_codegen.python.naming.function_name(
                aas_core_codegen.common.Identifier(f"{our_type.name}_from_str")
            )

            test_method_name = aas_core_codegen.python.naming.function_name(
                aas_core_codegen.common.Identifier(f"test_over_{descendant.name}")
            )

            test_methods.append(
                Stripped(
                    f"""\
def {test_method_name}(self) -> None:
{I}concrete_instance = (
{II}tests.common_jsonization.{load_concrete_maximal_name}()
{I})
{I}text = aas_xmlization.to_str(concrete_instance)
{I}et_concrete = ET.fromstring(text)

{I}# region From iterparse
{I}iterator = ET.iterparse(
{II}source=io.StringIO(text),
{II}events=['start', 'end']
{I})
{I}got_from_iterparse = (
{II}aas_xmlization.{from_iterparse}(
{III}iterator
{II})
{I})

{I}et_from_iterparse = (
{II}ET.fromstring(
{III}aas_xmlization.to_str(got_from_iterparse)
{II})
{I})
{I}tests.common_xmlization.remove_redundant_whitespace(
{II}et_from_iterparse
{I})
{I}tests.common_xmlization.assert_elements_equal(
{II}et_concrete, et_from_iterparse
{I})
{I}# endregion

{I}# region From stream
{I}got_from_stream = aas_xmlization.{from_stream}(
{II}io.StringIO(text)
{I})
{I}et_from_stream = (
{II}ET.fromstring(
{III}aas_xmlization.to_str(got_from_stream)
{II})
{I})
{I}tests.common_xmlization.remove_redundant_whitespace(
{II}et_from_stream
{I})
{I}tests.common_xmlization.assert_elements_equal(
{II}et_concrete, et_from_stream
{I})
{I}# endregion

{I}# region From file
{I}with tempfile.TemporaryDirectory() as tmp_dir:
{II}path = pathlib.Path(tmp_dir) / "something.xml"
{II}path.write_text(text, encoding='utf-8')

{II}got_from_file = aas_xmlization.{from_file}(
{III}path
{II})
{I}et_from_file = (
{II}ET.fromstring(
{III}aas_xmlization.to_str(got_from_file)
{II})
{I})
{I}tests.common_xmlization.remove_redundant_whitespace(
{II}et_from_file
{I})
{I}tests.common_xmlization.assert_elements_equal(
{II}et_concrete, et_from_file
{I})
{I}# endregion

{I}# region From string
{I}got_from_str = aas_xmlization.{from_str}(
{II}text
{I})
{I}et_from_str = (
{II}ET.fromstring(
{III}aas_xmlization.to_str(got_from_str)
{II})
{I})
{I}tests.common_xmlization.remove_redundant_whitespace(
{II}et_from_str
{I})
{I}tests.common_xmlization.assert_elements_equal(
{II}et_concrete, et_from_str
{I})
{I}# endregion"""
                )
            )

        class_writer = io.StringIO()
        class_writer.write(
            f"""\
class Test_{cls_name}(unittest.TestCase):
{I}\"\"\"
{I}Test round-trip XML de/serialization over concrete descendants
{I}of :py:class:`aas_types.{cls_name}`.
{I}\"\"\""""
        )

        for test_method in test_methods:
            class_writer.write("\n\n")
            class_writer.write(textwrap.indent(test_method, I))

        blocks.append(Stripped(class_writer.getvalue()))

    blocks.append(
        Stripped(
            """\
if __name__ == "__main__":
    unittest.main()"""
        )
    )

    blocks.append(warning)

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n\n")

        writer.write(block)

    writer.write("\n")

    target_pth = (
        dev_scripts.test_codegen.common.REPO_ROOT
        / "tests/test_xmlization_of_classes_with_descendants.py"
    )
    target_pth.write_text(writer.getvalue(), encoding="utf-8")

    return 0


if __name__ == "__main__":
    sys.exit(main())



"""Generate the common functions to de/serialize instances of a class."""
import enum
import io
import os
import pathlib
import sys
import textwrap
from typing import List, Optional

import aas_core_codegen.common
import aas_core_codegen.python.naming
import aas_core_codegen.naming
import aas_core_codegen.parse
import aas_core_codegen.run
from aas_core_codegen import intermediate
from aas_core_codegen.common import Stripped
import aas_core_codegen.python.common
from aas_core_codegen.python.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
    INDENT5 as IIIII,
    INDENT6 as IIIIII,
)

import dev_scripts.test_codegen.common
import dev_scripts.test_codegen.test_data_io


class ExampleKind(enum.Enum):
    """Represent kind of an instance example."""

    #: Maximal instance with all the properties set
    MAXIMAL = 0

    #: Minimal instance with only the required properties set
    MINIMAL = 1


def _generate_load_expected(
    cls: intermediate.ConcreteClass,
    container_cls: intermediate.ConcreteClass,
    example_kind: ExampleKind,
) -> Stripped:
    """
    Generate the function to load the maximal example from a container.

    If ``cls`` is self-contained, then ``cls is container_cls``.
    """

    # noinspection PyUnusedLocal
    container_kind = None  # type: Optional[Stripped]
    if cls is container_cls:
        container_kind = Stripped("SelfContained")
    elif container_cls.name == "Environment":
        container_kind = Stripped("ContainedInEnvironment")
    else:
        raise ValueError(f"Unexpected container class: {container_cls.name!r}")

    container_kind_literal = aas_core_codegen.python.common.string_literal(
        container_kind
    )

    model_type_literal = aas_core_codegen.python.common.string_literal(
        aas_core_codegen.naming.json_model_type(cls.name)
    )

    file_name_literal = None  # type: Optional[Stripped]
    if example_kind is ExampleKind.MAXIMAL:
        file_name_literal = Stripped(
            aas_core_codegen.python.common.string_literal("maximal.json")
        )
    elif example_kind is ExampleKind.MINIMAL:
        file_name_literal = Stripped(
            aas_core_codegen.python.common.string_literal("minimal.json")
        )
    else:
        aas_core_codegen.common.assert_never(example_kind)

    blocks = [
        Stripped(
            f"""\
path = (
{II}tests.common.TEST_DATA_DIR
{II}/ "Json"
{II}/ {container_kind_literal}
{II}/ "Expected"
{II}/ {model_type_literal}
{II}/ {file_name_literal}
)"""
        ),
        Stripped(
            f"""\
with path.open("rt") as fid:
{I}try:
{II}jsonable = json.load(fid)
{I}except json.decoder.JSONDecodeError as exception:
{II}raise AssertionError(
{III}f"Unexpected non-JSON content in {{path}}"
{II}) from exception"""
        ),
    ]  # type: List[Stripped]

    deserialization_function = aas_core_codegen.python.naming.function_name(
        aas_core_codegen.common.Identifier(f"{container_cls.name}_from_jsonable")
    )

    cls_name = aas_core_codegen.python.naming.class_name(cls.name)

    if cls is container_cls:
        blocks.append(
            Stripped(
                f"""\
return aas_jsonization.{deserialization_function}(
{I}jsonable
)"""
            )
        )
    else:
        blocks.append(
            Stripped(
                f"""\
container = aas_jsonization.{deserialization_function}(
{I}jsonable
)

instance: Optional[
{I}aas_types.{cls_name}
] = None

for something in container.descend():
{I}if isinstance(
{III}something,
{III}aas_types.{cls_name}
{I}):
{II}# We pick the least deep instance, so that we can also test for
{II}# nested instances *etc.*
{II}instance = something
{II}break

if instance is None:
{I}raise AssertionError(
{II}f"Expected to find an instance of {cls_name} "
{II}f"in {{path}}, but found none."
{I})

return instance"""
            )
        )

    function_name = None  # type: Optional[Stripped]
    if example_kind is ExampleKind.MAXIMAL:
        function_name = Stripped(
            aas_core_codegen.python.naming.function_name(
                aas_core_codegen.common.Identifier(f"load_maximal_{cls.name}")
            )
        )
    elif example_kind is ExampleKind.MINIMAL:
        function_name = Stripped(
            aas_core_codegen.python.naming.function_name(
                aas_core_codegen.common.Identifier(f"load_minimal_{cls.name}")
            )
        )
    else:
        aas_core_codegen.common.assert_never(example_kind)

    docstring_literal = None  # type: Optional[Stripped]
    if example_kind is ExampleKind.MAXIMAL:
        docstring_literal = Stripped(
            f"""\
\"\"\"
Load a maximal example
of :py:class:`aas_types.{cls_name}`
from :py:attr:`tests.common.TEST_DATA_DIR`.
\"\"\""""
        )
    elif example_kind is ExampleKind.MINIMAL:
        docstring_literal = Stripped(
            f"""\
\"\"\"
Load a minimal example
of :py:class:`aas_types.{cls_name}`
from :py:attr:`tests.common.TEST_DATA_DIR`.
\"\"\""""
        )
    else:
        aas_core_codegen.common.assert_never(example_kind)

    writer = io.StringIO()
    writer.write(
        f"""\
def {function_name}(
) -> aas_types.{cls_name}:
{I}{aas_core_codegen.common.indent_but_first_line(docstring_literal, I)}
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(textwrap.indent(block, I))

    return Stripped(writer.getvalue())


def main() -> int:
    """Execute the main routine."""
    symbol_table = dev_scripts.test_codegen.common.load_symbol_table()

    this_path = pathlib.Path(os.path.realpath(__file__))

    test_data_dir = dev_scripts.test_codegen.common.REPO_ROOT / "test_data"

    warning = dev_scripts.test_codegen.common.generate_warning_comment(
        this_path.relative_to(dev_scripts.test_codegen.common.REPO_ROOT)
    )

    blocks = [
        warning,
        Stripped('"""Provide common functionality for JSON de-serialization."""'),
        Stripped("# pylint: disable=missing-docstring"),
        Stripped(
            """\
import collections.abc
import json
import sys
from typing import Iterator, Optional

if sys.version_info >= (3, 8):
    from typing import Final
else:
    from typing_extensions import Final

import aas_core3.common as aas_common
import aas_core3.jsonization as aas_jsonization
import aas_core3.types as aas_types

import tests.common"""
        ),
        Stripped(
            f"""\
class Difference:
{I}\"\"\"Represent a single difference between two JSON-ables.\"\"\"

{I}#: Human-readable description of the difference
{I}message: Final[str]

{I}#: Path in the expected JSON-able value which is different from
{I}#: the obtained JSON-able value
{I}path: Final[aas_jsonization.Path]

{I}def __init__(self, message: str) -> None:
{II}\"\"\"Initialize with the given message and empty path.\"\"\"
{II}self.message = message
{II}self.path = aas_jsonization.Path()

{I}def __str__(self) -> str:
{II}return f"#{{self.path}}: {{self.message}}\""""
        ),
        Stripped(
            f"""\
def check_equal(
{I}expected: aas_jsonization.Jsonable,
{I}got: aas_jsonization.Jsonable,
) -> Iterator[Difference]:
{I}\"\"\"
{I}Compare recursively two JSON-able values for equality.

{I}:param expected: expected JSON-able value
{I}:param got: obtained JSON-able value
{I}:yield: differences
{I}\"\"\"
{I}if isinstance(expected, (bool, int, float, str, bytes)):
{II}if type(expected) != type(got):  # pylint: disable=unidiomatic-typecheck
{III}yield Difference(f"Expected {{type(expected)}}, but got {{type(got)}}")

{II}if expected != got:
{III}yield Difference(f"Expected {{expected!r}}, but got {{got!r}}")
{I}elif isinstance(expected, collections.abc.Sequence):
{II}if not isinstance(got, collections.abc.Sequence):
{III}yield Difference(f"Expected a sequence, but got {{type(got)}}")
{II}else:
{III}if len(expected) != len(got):
{IIII}yield Difference(
{IIIII}f"Expected a sequence of {{len(expected)}} item(s), "
{IIIII}f"but got {{len(got)}} item(s)"
{IIII})

{III}for i, (expected_item, got_item) in enumerate(zip(expected, got)):
{IIII}for difference in check_equal(expected_item, got_item):
{IIIII}difference.path._prepend(aas_jsonization.IndexSegment(expected, i))
{IIIII}yield difference

{I}elif isinstance(expected, collections.abc.Mapping):
{II}if not isinstance(got, collections.abc.Mapping):
{III}yield Difference(f"Expected a mapping, but got {{type(got)}}")
{II}else:
{III}if not all(isinstance(key, str) for key in expected.keys()):
{IIII}raise ValueError(
{IIIII}f"Expected all keys in the expected JSON-able value to be strings, "
{IIIII}f"but got: {{list(expected.keys())}}"
{IIII})

{III}if not all(isinstance(key, str) for key in got.keys()):
{IIII}yield Difference(
{IIIII}f"Expected all keys in a mapping to be strings, "
{IIIII}f"but got: {{list(got.keys())}}"
{IIII})

{III}expected_key_set = set(expected.keys())
{III}got_key_set = set(got.keys())

{III}expected_got_diff = expected_key_set.difference(got_key_set)
{III}if expected_got_diff:
{IIII}yield Difference(f"Expected key(s) {{sorted(expected_got_diff)}} missing")

{III}got_expected_diff = got_key_set.difference(expected_key_set)
{III}if got_expected_diff:
{IIII}yield Difference(f"Unexpected key(s) {{sorted(got_expected_diff)}}")

{III}for key, expected_value in expected.items():
{IIII}got_value = got[key]

{IIII}for difference in check_equal(expected_value, got_value):
{IIIII}difference.path._prepend(
{IIIIII}aas_jsonization.PropertySegment(expected, key)
{IIIII})
{IIIII}yield difference
{I}else:
{II}aas_common.assert_never(expected)"""
        ),
    ]  # type: List[Stripped]

    environment_cls = symbol_table.must_find_concrete_class(
        aas_core_codegen.common.Identifier("Environment")
    )

    for our_type in symbol_table.our_types:
        if not isinstance(our_type, intermediate.ConcreteClass):
            continue

        container_cls = dev_scripts.test_codegen.test_data_io.determine_container_class(
            cls=our_type, test_data_dir=test_data_dir, environment_cls=environment_cls
        )

        blocks.append(
            _generate_load_expected(
                cls=our_type,
                container_cls=container_cls,
                example_kind=ExampleKind.MAXIMAL,
            )
        )

        blocks.append(
            _generate_load_expected(
                cls=our_type,
                container_cls=container_cls,
                example_kind=ExampleKind.MINIMAL,
            )
        )

    blocks.append(warning)

    writer = io.StringIO()

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n\n")

        writer.write(block)

    writer.write("\n")

    target_pth = (
        dev_scripts.test_codegen.common.REPO_ROOT / "tests/common_jsonization.py"
    )
    target_pth.write_text(writer.getvalue(), encoding="utf-8")

    return 0


if __name__ == "__main__":
    sys.exit(main())
